import math
import re

import pandas as pd
from ete3 import Tree

from genomics_data_index.api.query.GenomicsDataIndex import GenomicsDataIndex
from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl.ClusterScorer import ClusterScorer
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.model.db import Sample


# wrapper methods to simplify writing tests
def query(connection: DataIndexConnection, **kwargs) -> SamplesQuery:
    return GenomicsDataIndex(connection=connection).samples_query(**kwargs)


def test_score_samples_with_mutation_tree(prebuilt_tree: Tree, loaded_database_connection: DataIndexConnection):
    query_tree = query(loaded_database_connection).join_tree(tree=prebuilt_tree, kind='mutation',
                                                             reference_name='genome',
                                                             alignment_length=5180)
    assert 3 == len(query_tree)
    assert 9 == len(query_tree.universe_set)

    # Test single sample
    query_result = query_tree.isa('SampleA')
    assert 1 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isclose(1 / 1, cluster_scorer.score_samples(query_result), rel_tol=1e-3)

    # Test single sample set
    query_result = query_tree.isa('SampleA')
    assert 1 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isclose(1 / 1, cluster_scorer.score_samples(query_result.sample_set), rel_tol=1e-3)

    # Test multiple samples
    query_result = query_tree.isin(['SampleB', 'SampleC'])
    assert 2 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isclose(2 / 2, cluster_scorer.score_samples(query_result), rel_tol=1e-3)

    # Test multiple samples set
    query_result = query_tree.isin(['SampleB', 'SampleC'])
    assert 2 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isclose(2 / 2, cluster_scorer.score_samples(query_result.sample_set), rel_tol=1e-3)

    # Test multiple samples non-perfect score
    query_result = query_tree.isin(['SampleA', 'SampleB'])
    assert 2 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isclose(2 / 3, cluster_scorer.score_samples(query_result), rel_tol=1e-3)

    # Test multiple samples set non-perfect score
    query_result = query_tree.isin(['SampleA', 'SampleB'])
    assert 2 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isclose(2 / 3, cluster_scorer.score_samples(query_result.sample_set), rel_tol=1e-3)

    # Test multiple samples non-perfect score 2
    query_result = query_tree.isin(['SampleA', 'SampleC'])
    assert 2 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isclose(2 / 3, cluster_scorer.score_samples(query_result), rel_tol=1e-3)

    # Test all samples
    query_result = query_tree.isin(['SampleA', 'SampleB', 'SampleC'])
    assert 3 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isclose(3 / 3, cluster_scorer.score_samples(query_result), rel_tol=1e-3)

    # # Test no samples
    query_result = query_tree.isa('not exists')
    assert 0 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isnan(cluster_scorer.score_samples(query_result))

    # # Test no samples set
    query_result = query_tree.isa('not exists')
    assert 0 == len(query_result)
    cluster_scorer = ClusterScorer.create(query_tree)
    assert math.isnan(cluster_scorer.score_samples(query_result.sample_set))


def test_score_groupby_with_mutation_tree(prebuilt_tree: Tree, loaded_database_connection: DataIndexConnection):
    db = loaded_database_connection.database
    sampleA = db.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = db.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = db.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    metadata_df = pd.DataFrame([
        [sampleA.id, 'red', 'up', 'A', '1', 'red', 'A'],
        [sampleB.id, 'blue', 'up', 'B', '1', 'red', 'A.1'],
        [sampleC.id, 'blue', 'down', 'C', '1', pd.NA, 'A.1.1']
    ], columns=['Sample ID', 'Color', 'Direction', 'Letter', 'Number', 'Color_NA', 'Type'])

    query_tree = query(loaded_database_connection).join_tree(tree=prebuilt_tree, kind='mutation',
                                                             reference_name='genome',
                                                             alignment_length=5180)
    query_tree_df = query_tree.join(metadata_df, sample_ids_column='Sample ID')
    cluster_scorer = ClusterScorer.create(query_tree_df)

    # Group by 2 groups perfect score
    scores_df = cluster_scorer.score_groupby('Color')
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 2 == len(scores_df)
    assert {'red', 'blue'} == set(scores_df.index)
    assert math.isclose(1 / 1, scores_df.loc['red', 'Score'], abs_tol=1e-3)
    assert math.isclose(2 / 2, scores_df.loc['blue', 'Score'], abs_tol=1e-3)

    # Group by 2 groups imperfect score
    scores_df = cluster_scorer.score_groupby('Direction')
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 2 == len(scores_df)
    assert {'up', 'down'} == set(scores_df.index)
    assert math.isclose(2 / 3, scores_df.loc['up', 'Score'], abs_tol=1e-3)
    assert math.isclose(1 / 1, scores_df.loc['down', 'Score'], abs_tol=1e-3)

    # Group by 3 groups
    scores_df = cluster_scorer.score_groupby('Letter')
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 3 == len(scores_df)
    assert {'A', 'B', 'C'} == set(scores_df.index)
    assert math.isclose(1 / 1, scores_df.loc['A', 'Score'], abs_tol=1e-3)
    assert math.isclose(1 / 1, scores_df.loc['B', 'Score'], abs_tol=1e-3)
    assert math.isclose(1 / 1, scores_df.loc['C', 'Score'], abs_tol=1e-3)

    # Group by 1 group
    scores_df = cluster_scorer.score_groupby('Number')
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 1 == len(scores_df)
    assert {'1'} == set(scores_df.index)
    assert math.isclose(1 / 1, scores_df.loc['1', 'Score'], abs_tol=1e-3)

    # Group by with NA exclude NA
    scores_df = cluster_scorer.score_groupby('Color_NA')
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 1 == len(scores_df)
    assert {'red'} == set(scores_df.index)
    assert math.isclose(2 / 3, scores_df.loc['red', 'Score'], abs_tol=1e-3)

    # Group by with NA, fill na
    scores_df = cluster_scorer.score_groupby('Color_NA', na_value='NA')
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 2 == len(scores_df)
    assert {'red', 'NA'} == set(scores_df.index)
    assert math.isclose(2 / 3, scores_df.loc['red', 'Score'], abs_tol=1e-3)
    assert math.isclose(1 / 1, scores_df.loc['NA', 'Score'], abs_tol=1e-3)

    # Group by 2 groups exclude 1 due to minimum samples
    scores_df = cluster_scorer.score_groupby('Color', min_samples_count=2)
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 1 == len(scores_df)
    assert {'blue'} == set(scores_df.index)
    assert math.isclose(2 / 2, scores_df.loc['blue', 'Score'], abs_tol=1e-3)

    # Group by 2 groups exclude 1 due to maximum samples
    scores_df = cluster_scorer.score_groupby('Color', max_samples_count=1)
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 1 == len(scores_df)
    assert {'red'} == set(scores_df.index)
    assert math.isclose(1 / 1, scores_df.loc['red', 'Score'], abs_tol=1e-3)

    # Group by 3 groups using Type
    scores_df = cluster_scorer.score_groupby('Type')
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 3 == len(scores_df)
    assert {'A', 'A.1', 'A.1.1'} == set(scores_df.index)
    assert math.isclose(1 / 1, scores_df.loc['A', 'Score'], abs_tol=1e-3)
    assert math.isclose(1 / 1, scores_df.loc['A.1', 'Score'], abs_tol=1e-3)
    assert math.isclose(1 / 1, scores_df.loc['A.1.1', 'Score'], abs_tol=1e-3)

    # Group by A.1 group using Type and groupby_func
    scores_df = cluster_scorer.score_groupby('Type', groupby_func=
    lambda x: m.group(1) if (m := re.search(r'^([^.]+\.[^.]+)', x)) else pd.NA)
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 1 == len(scores_df)
    assert {'A.1'} == set(scores_df.index)
    assert math.isclose(2 / 2, scores_df.loc['A.1', 'Score'], abs_tol=1e-3)

    # Group A.1 group using Type and groupby_func
    scores_df = cluster_scorer.score_groupby('Type', groupby_func=
    lambda x: m.group(1) if (m := re.search(r'^([^.]+)', x)) else pd.NA)
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 1 == len(scores_df)
    assert {'A'} == set(scores_df.index)
    assert math.isclose(3 / 3, scores_df.loc['A', 'Score'], abs_tol=1e-3)

    # Group by 2 groups using Type and groupby_func
    scores_df = cluster_scorer.score_groupby('Type', groupby_func=
    lambda x: m.group(1) if (m := re.search(r'^([^.]+\.[^.]+)', x)) else x)
    assert ['Score', 'Sample Count'] == scores_df.columns.tolist()
    assert 2 == len(scores_df)
    assert {'A', 'A.1'} == set(scores_df.index)
    assert math.isclose(1 / 1, scores_df.loc['A', 'Score'], abs_tol=1e-3)
    assert math.isclose(2 / 2, scores_df.loc['A.1', 'Score'], abs_tol=1e-3)



import pytest
from ete3 import Tree

from genomics_data_index.test.integration import tree_file


@pytest.fixture
def expected_tree() -> Tree:
    return Tree(str(tree_file))


def test_build_tree_core_fasttree(tree_service, core_alignment_service, expected_tree):
    alignment = core_alignment_service.construct_alignment(
        reference_name='genome', samples=['SampleA', 'SampleB', 'SampleC'],
        include_variants=['SNP'], align_type='core',
    )

    tree, out = tree_service.build_tree(alignment, tree_build_type='fasttree', )

    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(tree.get_leaf_names())

    tree_comparison = expected_tree.compare(tree, unrooted=True)
    assert tree_comparison['rf'] == 0


def test_build_tree_core_iqtree(tree_service, core_alignment_service, expected_tree):
    alignment = core_alignment_service.construct_alignment(
        reference_name='genome', samples=['SampleA', 'SampleB', 'SampleC'],
        include_variants=['SNP'], align_type='core',
    )

    tree, out = tree_service.build_tree(alignment, tree_build_type='iqtree',
                                        extra_params='--seed 42 -m GTR+ASC')

    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(tree.get_leaf_names())

    tree_comparison = expected_tree.compare(tree, unrooted=True)
    assert tree_comparison['rf'] == 0

    actual_distance = 5180 * tree.get_distance('SampleA', 'genome')
    expected_distance = 26
    assert abs(expected_distance - actual_distance) < 6


def test_build_tree_core_iqtree_2cores(tree_service, core_alignment_service, expected_tree):
    alignment = core_alignment_service.construct_alignment(
        reference_name='genome', samples=['SampleA', 'SampleB', 'SampleC'],
        include_variants=['SNP'], align_type='core',
    )

    tree, out = tree_service.build_tree(alignment, tree_build_type='iqtree', num_cores=2)

    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(tree.get_leaf_names())

    tree_comparison = expected_tree.compare(tree, unrooted=True)
    assert tree_comparison['rf'] == 0


def test_build_tree_full(tree_service, core_alignment_service, expected_tree):
    alignment = core_alignment_service.construct_alignment(
        reference_name='genome', samples=['SampleA', 'SampleB', 'SampleC'], align_type='full')

    tree, out = tree_service.build_tree(alignment, tree_build_type='iqtree',
                                        extra_params='--seed 42 -m GTR')

    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(tree.get_leaf_names())

    tree_comparison = expected_tree.compare(tree, unrooted=True)
    assert tree_comparison['rf'] == 0

    actual_distance = 5180 * tree.get_distance('SampleA', 'genome')
    expected_distance = 26
    assert abs(expected_distance - actual_distance) < 6


def test_build_tree_two_samples(tree_service, core_alignment_service):
    alignment = core_alignment_service.construct_alignment(
        reference_name='genome', samples=['SampleA', 'SampleB'], align_type='full')

    tree, out = tree_service.build_tree(alignment)

    assert {'SampleA', 'SampleB', 'genome'} == set(tree.get_leaf_names())


def test_rebuild_tree(tree_service, reference_service_with_data, expected_tree):
    tree_service.rebuild_tree(reference_name='genome', align_type='core', include_variants=['SNP'])
    reference_genome = reference_service_with_data.find_reference_genome('genome')
    tree = reference_genome.tree

    assert 58 == reference_genome.tree_alignment_length
    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(tree.get_leaf_names())

    tree_comparison = expected_tree.compare(tree, unrooted=True)
    assert tree_comparison['rf'] == 0



from os import path, listdir
from pathlib import Path
from typing import List

root_data_dir = Path(path.dirname(__file__), 'data')
mlst_dir = root_data_dir / 'mlst'
data_dir = root_data_dir / 'snippy'
snpeff_data_dir = root_data_dir / 'snpeff-vcf'
snpeff_database_dir = root_data_dir / 'snpeff-database'
extra_snippy_dir = root_data_dir / 'extra-snippy-samples'
regular_vcf_dir = root_data_dir / 'regular-vcf'
variation_dir = root_data_dir / 'variation'
consensus_dir = root_data_dir / 'consensus'
sourmash_dir = root_data_dir / 'sourmash'
sample_dirs = [data_dir / d for d in listdir(data_dir) if path.isdir(data_dir / d)]
masked_positions_snippy_bed = data_dir / 'missing-all.bed.gz'
sample_dirs_AB = [data_dir / 'SampleA', data_dir / 'SampleB']
sample_dirs_C = [data_dir / 'SampleC']
sample_dirs_A = [data_dir / 'SampleA']
sample_dirs_BC = [data_dir / 'SampleB', data_dir / 'SampleC']
reference_file = data_dir / 'genome.fasta.gz'
reference_file_lzma = data_dir / 'genome.fasta.xz'
reference_file_bzip2 = data_dir / 'genome.fasta.bz2'
reference_file_snpeff = snpeff_data_dir / 'NC_011083.fasta.gz'
reference_file_5000_snpeff = snpeff_database_dir / 'NC_011083-5000.gbk.gz'
reference_file_5000_snpeff_2 = snpeff_database_dir / 'NC_011083_CP001602-5000.gbk.gz'
snpeff_vcf_file = snpeff_database_dir / 'SampleA.vcf.gz'
tree_file = data_dir / 'tree.tre'
snpeff_tree_file = snpeff_data_dir / 'tree.txt'
basic_mlst_file = mlst_dir / 'mlst-basic.tsv'
mlst_file_unknown = mlst_dir / 'mlst-unknown.tsv'
sistr_mlst_file = mlst_dir / 'mlst-sistr.csv'
chewbbaca_mlst_file = mlst_dir / 'mlst-chewbbaca-small.tsv'
mlst_file_single_scheme = mlst_dir / 'mlst-single-scheme.tsv'
mlst_file_single_scheme2 = mlst_dir / 'mlst-single-scheme2.tsv'
mlst_file_single_scheme3 = mlst_dir / 'mlst-single-scheme3.tsv'
mlst_snippy_file = mlst_dir / 'mlst-snippy-data.csv'

data_dir_empty = root_data_dir / 'empty_vcfs'

test_project_dir = root_data_dir / 'test_project_dir'

snippy_sample_vcfs_dict = {
    'SampleA': data_dir / 'SampleA' / 'snps.vcf.gz',
    'SampleB': data_dir / 'SampleB' / 'snps.vcf.gz',
    'SampleC': data_dir / 'SampleC' / 'snps.vcf.gz',
}

snippy_sample_mask_sequences_dict = {
    'SampleA': data_dir / 'SampleA' / 'snps.aligned.fa',
    'SampleB': data_dir / 'SampleB' / 'snps.aligned.fa',
    'SampleC': data_dir / 'SampleC' / 'snps.aligned.fa',
}

snippy_sample2_vcfs_dict = {
    'SampleA_2': data_dir / 'SampleA' / 'snps.vcf.gz',
    'SampleB_2': data_dir / 'SampleB' / 'snps.vcf.gz',
    'SampleC_2': data_dir / 'SampleC' / 'snps.vcf.gz',
}

snippy_sample2_mask_sequences_dict = {
    'SampleA_2': data_dir / 'SampleA' / 'snps.aligned.fa',
    'SampleB_2': data_dir / 'SampleB' / 'snps.aligned.fa',
    'SampleC_2': data_dir / 'SampleC' / 'snps.aligned.fa',
}

sourmash_signatures = {
    'SampleA': sourmash_dir / 'SampleA.sig.gz',
    'SampleB': sourmash_dir / 'SampleB.sig.gz',
    'SampleC': sourmash_dir / 'SampleC.sig.gz',
}

snippy_snps_dataframes = {
    'SampleA': data_dir / 'SampleA' / 'mutations-dataframe.snps.tsv',
    'SampleB': data_dir / 'SampleB' / 'mutations-dataframe.snps.tsv',
    'SampleC': data_dir / 'SampleC' / 'mutations-dataframe.snps.tsv',
}

snippy_all_dataframes = {
    'SampleA': data_dir / 'SampleA' / 'mutations-dataframe.all.tsv',
    'SampleB': data_dir / 'SampleB' / 'mutations-dataframe.all.tsv',
    'SampleC': data_dir / 'SampleC' / 'mutations-dataframe.all.tsv',
}

snpeff_sample_vcfs = {
    'SH10-014': snpeff_data_dir / 'SH10-014.vcf.gz',
    'SH14-001': snpeff_data_dir / 'SH14-001.vcf.gz',
    'SH14-014': snpeff_data_dir / 'SH14-014.vcf.gz',
}

snpeff_sample_beds = {
    'SH10-014': snpeff_data_dir / 'SH10-014.bed.gz',
    'SH14-001': snpeff_data_dir / 'SH14-001.bed.gz',
    'SH14-014': snpeff_data_dir / 'SH14-014.bed.gz',
}

snpeff_sarscov2_vcfs = {
    'USA/CA-CDPH-3000143037/2021': snpeff_data_dir / 'USA__CA-CDPH-3000143037__2021.vcf.gz',
    'USA/CA-CDPH-3000143037/2021.noann': snpeff_data_dir / 'USA__CA-CDPH-3000143037__2021.noann.vcf.gz'
}

snpeff_sample_vcfs_fake_dup = {
    'SH10-014-dup-gene-variant': snpeff_data_dir / 'SH10-014-dup-gene-variant.vcf.gz',
    'SH10-014-dup-gene-variant-2': snpeff_data_dir / 'SH10-014-dup-gene-variant-2.vcf.gz',
}


def expand_list_by(list_in: List[str], number: int) -> List[str]:
    new_list = []
    for value in list_in:
        new_list.extend([value] * number)

    return new_list



import abc
from typing import Generator, List

import pandas as pd

from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData


class VcfVariantsTableProcessor(abc.ABC):

    def __init__(self, include_masked_regions: bool):
        self._include_masked_regions = include_masked_regions

    @property
    def include_masked_regions(self) -> bool:
        return self._include_masked_regions

    @abc.abstractmethod
    def process(self, sample_data: List[NucleotideSampleData]) -> Generator[pd.DataFrame, None, None]:
        pass


class VcfVariantsTableProcessorFactory(abc.ABC):

    def __init__(self):
        pass

    @abc.abstractmethod
    def create_variants_processor(self, include_masked_regions: bool) -> VcfVariantsTableProcessor:
        """
        Builds a new instance of a VcfVariantsTableProcessor.
        The reason I have a parallel factory class for each VcfVariantsTableProcessor is that for multiprocessing
        I cannot set a separate variable 'include_masked_regions' before I use Python's multiprocessing pool
        to divide up tasks. Or, I can set such a variable, but it has to be a class variable. But the decision
        on whether or not to include masked regions needs to be made much later than the decision on whether or
        not I want to use parallel/serial processing. So I create a Factory which defines the decision on whether to
        use parallel/serial processing and then the moment the decision on whether or not to use 'include_masked_regions'
        is made I can use a factory to create a new instance of VcfVariantsTableProcessor with the appropriate class
        variable set.
        :param include_masked_regions: Whether or not masked/missing/unknown regions should be included.
        :return: A new instance of a VcfVariantsTableProcessor.
        """
        pass



import gzip

import pytest
from Bio import SeqIO
from Bio.Seq import Seq
from ete3 import Tree
from sqlalchemy.orm.exc import NoResultFound

from genomics_data_index.storage.model.db import Reference
from genomics_data_index.storage.service import EntityExistsError
from genomics_data_index.test.integration import reference_file
from genomics_data_index.test.integration import tree_file


@pytest.fixture
def example_tree() -> Tree:
    return Tree(str(tree_file))


def test_count_references(reference_service):
    assert 0 == reference_service.count_reference_genomes()


def count_references_with_data(reference_service_with_data):
    assert 1 == reference_service_with_data.count_reference_genomes()


def test_translate_spdi(reference_service_with_data):
    spdi_ids = {
        'reference:3762:1:G',
        'reference:374:1:ATTCTAGGGTAGACGCT',
        'reference:3576:4:T',
        'reference:3513:1:G',
        'reference:1984:7:TTGA',
        'reference:1483:21:A',
    }

    translated_ids = reference_service_with_data.translate_spdi(spdi_ids, to='spdi_ref')
    assert translated_ids['reference:3762:1:G'] == 'reference:3762:A:G'
    assert translated_ids['reference:374:1:ATTCTAGGGTAGACGCT'] == 'reference:374:A:ATTCTAGGGTAGACGCT'
    assert translated_ids['reference:3576:4:T'] == 'reference:3576:TTTC:T'
    assert translated_ids['reference:3513:1:G'] == 'reference:3513:T:G'
    assert translated_ids['reference:1984:7:TTGA'] == 'reference:1984:GTGATTG:TTGA'
    assert translated_ids['reference:1483:21:A'] == 'reference:1483:AAAGAGGGGCTGCTGGAGCCG:A'

    spdi_ids_ref = {
        'reference:3762:A:G',
        'reference:374:A:ATTCTAGGGTAGACGCT',
        'reference:3576:TTTC:T',
        'reference:3513:T:G',
        'reference:1984:GTGATTG:TTGA',
        'reference:1483:AAAGAGGGGCTGCTGGAGCCG:A'
    }

    translated_ids = reference_service_with_data.translate_spdi(spdi_ids_ref, to='spdi_ref')
    assert translated_ids['reference:3762:A:G'] == 'reference:3762:A:G'
    assert translated_ids['reference:374:A:ATTCTAGGGTAGACGCT'] == 'reference:374:A:ATTCTAGGGTAGACGCT'
    assert translated_ids['reference:3576:TTTC:T'] == 'reference:3576:TTTC:T'
    assert translated_ids['reference:3513:T:G'] == 'reference:3513:T:G'
    assert translated_ids['reference:1984:GTGATTG:TTGA'] == 'reference:1984:GTGATTG:TTGA'
    assert translated_ids['reference:1483:AAAGAGGGGCTGCTGGAGCCG:A'] == 'reference:1483:AAAGAGGGGCTGCTGGAGCCG:A'


def test_insert_reference_genome(database, reference_service):
    assert 0 == database.get_session().query(Reference).count(), 'Database should be empty initially'
    reference_service.add_reference_genome(reference_file)
    assert 1 == database.get_session().query(Reference).count(), 'Database should have one entry'
    assert 'genome' == database.get_session().query(Reference).all()[0].name, 'Name should match'

    seq_record = reference_service.get_sequence('reference')

    assert seq_record is not None, 'No matching seq record'
    assert seq_record.id == 'reference', 'Incorrect record id'

    with gzip.open(reference_file, mode='rt', encoding='ascii') as f:
        records = list(SeqIO.parse(f, 'fasta'))
        assert records[0].seq == seq_record.seq, 'Incorrect sequence'


def test_double_insert_reference_genome(database, reference_service):
    assert 0 == database.get_session().query(Reference).count(), 'Database should be empty initially'
    reference_service.add_reference_genome(reference_file)
    assert 1 == database.get_session().query(Reference).count(), 'Database should have one entry'
    assert 'genome' == database.get_session().query(Reference).all()[0].name, 'Name should match'

    with pytest.raises(EntityExistsError) as execinfo:
        reference_service.add_reference_genome(reference_file)
    assert 'Reference genome [genome] already exists in database' in str(execinfo.value)
    assert 1 == database.get_session().query(Reference).count(), 'Database should have one entry'


def test_find_reference_genome(database, reference_service):
    assert 0 == database.get_session().query(Reference).count(), 'Database should be empty initially'
    reference_service.add_reference_genome(reference_file)

    reference = reference_service.find_reference_genome('genome')
    assert 'genome' == reference.name, 'Reference name should match'
    assert 5180 == reference.length, 'Reference length should match'


def test_get_nonexistent_sequence(reference_service):
    with pytest.raises(KeyError) as execinfo:
        reference_service.get_sequence('does_not_exist')
    assert 'Alias does_not_exist' in str(execinfo.value)


def test_get_reference_genomes(reference_service):
    reference_service.add_reference_genome(reference_file)
    assert {'genome'} == {genome.name for genome in reference_service.get_reference_genomes()}


def test_get_reference_genomes_empty(reference_service):
    assert [] == reference_service.get_reference_genomes()


def test_update_tree(database, reference_service, example_tree):
    reference_service.add_reference_genome(reference_file)

    reference_genome = database.get_session().query(Reference).filter(Reference.name == 'genome').one()
    assert reference_genome is not None
    with pytest.raises(Exception) as execinfo:
        reference_genome.tree
    assert 'Cannot convert an empty tree' in str(execinfo.value)

    reference_service.update_tree(reference_name='genome', tree=example_tree, alignment_length=1000)
    reference_genome = database.get_session().query(Reference).filter(Reference.name == 'genome').one()
    assert reference_genome.tree.write() == example_tree.write()
    assert 1000 == reference_genome.tree_alignment_length


def test_find_references_for_sample(reference_service_with_data, variation_service):
    found_references = reference_service_with_data.find_references_for_sample('SampleA')
    assert len(found_references) == 1
    assert 'genome' == found_references[0].name

    found_references = reference_service_with_data.find_references_for_sample('SampleB')
    assert len(found_references) == 1
    assert 'genome' == found_references[0].name

    found_references = reference_service_with_data.find_references_for_sample('SampleC')
    assert len(found_references) == 1
    assert 'genome' == found_references[0].name


def test_find_references_for_sample_not_exist(reference_service_with_data, variation_service):
    found_references = reference_service_with_data.find_references_for_sample('not_exist')
    assert len(found_references) == 0


def test_get_reference_sequences(reference_service_with_data, variation_service):
    reference_sequences = reference_service_with_data.get_reference_sequences('genome')

    assert {'reference'} == set(reference_sequences.keys())


def test_find_reference_for_sequence(reference_service_with_data, variation_service):
    reference = reference_service_with_data.find_reference_for_sequence('reference')
    assert 'genome' == reference.name


def test_find_reference_for_sequence_not_exist(reference_service_with_data, variation_service):
    with pytest.raises(NoResultFound) as execinfo:
        reference_service_with_data.find_reference_for_sequence('not_exist')
    assert 'No row was found' in str(execinfo.value)


def test_get_sequence(reference_service_with_data):
    seq_record = reference_service_with_data.get_sequence('reference')

    assert 'reference' == seq_record.id
    assert isinstance(seq_record.seq, Seq)


def test_get_reference_genome_records(reference_service_with_data):
    records = reference_service_with_data.get_reference_genome_records('genome')
    assert 1 == len(records)

    assert 'reference' == records[0].id
    assert 5180 == len(records[0])



from typing import Dict, Set, Any


def get_values_from_signatures(sigs) -> Dict[str, Set[Any]]:
    ksizes = set()
    scales = set()

    for sig in sigs:
        ksizes.add(sig.minhash.ksize)
        scales.add(sig.minhash.scaled)

    return {
        'ksize': ksizes,
        'scaled': scales,
    }



import logging

from ete3 import Tree, TreeStyle

from genomics_data_index.api.viewer.samples_visuals.AbstractAnnotationTreeSamplesVisual import \
    AbstractAnnotationTreeSamplesVisual

logger = logging.getLogger(__name__)


class AnnotationSpacingTreeSamplesVisual(AbstractAnnotationTreeSamplesVisual):
    """
    Creates a visual style which adds empty spaces in annotation table.
    """

    def __init__(self, width: int, height: int, annotate_column: int, color: str = None,
                 border_width: int = None, border_color: str = None):
        super().__init__(samples=set(),
                         legend_nodesize=0,
                         legend_fontsize=0,
                         legend_columns={},
                         annotate_kind='r',
                         annotate_border_width=border_width,
                         annotate_margin=0)
        self._box_width = width
        self._box_height = height
        self._annotate_column = annotate_column
        self._color = color
        self._annotation_opacity = 0.0
        self._annotate_border_color = border_color

    def apply_visual(self, tree: Tree, tree_style: TreeStyle) -> None:
        for leaf in tree.iter_leaves():
            annotate_face = self._build_annotate_face(width=self._box_width, height=self._box_height,
                                                      border_color=self._annotate_border_color,
                                                      bgcolor=self._color,
                                                      opacity=self._annotation_opacity,
                                                      label=None)

            leaf.add_face(annotate_face, column=self._annotate_column, position='aligned')



#!/usr/bin/env python
# Used to print out the nucleotide and position in the reference genome to fill in information for tests.

import gzip
import sys
from functools import partial

from Bio import SeqIO

if len(sys.argv) != 2:
    print(f"Usage: {sys.argv[0]} [reference.fasta]\n")
    sys.exit(1)
file = sys.argv[1]

_open = partial(gzip.open, mode='rt') if file.endswith('.gz') else open

with _open(file) as handle:
    for record in SeqIO.parse(handle, "fasta"):
        name = record.id
        for i, c in enumerate(record.seq):
            c = c.upper()
            pos = i + 1
            print(f'{pos}\t{c}')



import pytest

from genomics_data_index.storage.io.mlst.MLSTChewbbacaReader import MLSTChewbbacaReader
from genomics_data_index.test.integration import chewbbaca_mlst_file, expand_list_by


@pytest.fixture
def mlst_reader() -> MLSTChewbbacaReader:
    return MLSTChewbbacaReader(mlst_file=chewbbaca_mlst_file, scheme='test-scheme')


def test_get_features_table(mlst_reader):
    num_samples = 2
    num_loci = 11

    mlst_df = mlst_reader.get_features_table()

    assert ['File', 'Sample', 'Scheme', 'Locus', 'Allele'] == list(mlst_df.columns)
    assert {'test-scheme'} == set(mlst_df['Scheme'].tolist())

    assert num_samples * num_loci == len(mlst_df)
    assert expand_list_by(['GCF_000006945.2_ASM694v2_genomic.fna', 'GCF_000007545.1_ASM754v1_genomic.fna'],
                          num_loci) == list(mlst_df['File'].tolist())
    assert expand_list_by(['GCF_000006945', 'GCF_000007545'],
                          num_loci) == list(mlst_df['Sample'].tolist())

    assert ['1'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000006945') &
                                     (mlst_df['Locus'] == 'GCF-000006945-protein10.fasta'), 'Allele'].tolist())

    assert ['2'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000007545') &
                                     (mlst_df['Locus'] == 'GCF-000006945-protein1004.fasta'), 'Allele'].tolist())


def test_get_features_table_unknown(mlst_reader):
    mlst_df = mlst_reader.get_features_table()

    assert ['File', 'Sample', 'Scheme', 'Locus', 'Allele'] == list(mlst_df.columns)
    assert {'test-scheme'} == set(mlst_df['Scheme'].tolist())

    assert ['?'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000007545') &
                                     (mlst_df['Locus'] == 'GCF-000006945-protein1.fasta'), 'Allele'].tolist())

    assert ['?'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000007545') &
                                     (mlst_df['Locus'] == 'GCF-000006945-protein1008.fasta'), 'Allele'].tolist())



from __future__ import annotations

from typing import Optional, List, Tuple

from genomics_data_index.storage.model.QueryFeature import QueryFeature
from genomics_data_index.storage.model.QueryFeatureMutation import QueryFeatureMutation


class QueryFeatureHGVS(QueryFeatureMutation):
    PREFIX = 'hgvs:'

    def __init__(self, sequence_name: str, gene: Optional[str], variant: str):
        super().__init__()

        self._sequence_name = sequence_name
        self._gene = gene
        self._variant = variant

        if self._variant is not None:
            if self._gene is None:
                self._hgvs_id = f'{sequence_name}:{variant}'
            else:
                self._hgvs_id = f'{sequence_name}:{gene}:{variant}'
        else:
            self._hgvs_id = None

    def has_gene(self) -> bool:
        return self._gene is not None

    def has_id(self) -> bool:
        return self._hgvs_id is not None

    def is_nucleotide(self) -> bool:
        return self.has_id() and (self._variant.startswith('n.') or self._variant.startswith('c.'))

    def is_protein(self) -> bool:
        return self.has_id() and self._variant.startswith('p.')

    @property
    def prefix(self) -> str:
        return self.PREFIX

    @property
    def sequence(self) -> str:
        return self._sequence_name

    @property
    def gene(self) -> str:
        return self._gene

    @property
    def variant(self) -> str:
        return self._variant

    @property
    def id(self) -> Optional[str]:
        if self.has_id():
            return super().id
        else:
            return None

    @property
    def id_no_prefix(self) -> Optional[str]:
        return self._hgvs_id

    def is_unknown(self) -> bool:
        raise NotImplementedError('Not implemented')

    @property
    def scope(self) -> str:
        return self.sequence

    def to_unknown(self) -> QueryFeature:
        raise NotImplementedError('Not implemented')

    def to_unknown_explode(self) -> List[QueryFeature]:
        raise NotImplementedError('Not implemented')

    @classmethod
    def create(cls, sequence_name: str, gene: Optional[str], variant: Optional[str]) -> QueryFeatureHGVS:
        # If variant is given as nucleotide coordinates with respect to the sequence 'n.' then ignore gene id
        if variant is not None and variant.startswith('n.'):
            gene = None

        return QueryFeatureHGVS(sequence_name=sequence_name, gene=gene, variant=variant)

    @classmethod
    def split_id(cls, hgvs_id: str, prefix: str = None) -> Tuple[str, str, str]:
        if prefix is None:
            prefix = cls.PREFIX

        if hgvs_id.startswith(f'{prefix}'):
            hgvs_id_strip = hgvs_id[len(f'{prefix}'):]
        else:
            raise Exception(f'hgvs_id=[{hgvs_id}] must start with [{prefix}]')

        tokens = hgvs_id_strip.split(cls.SPLIT_CHAR)
        if len(tokens) == 2:
            sequence_name = tokens[0]
            gene = None
            variant = tokens[1]
        elif len(tokens) == 3:
            sequence_name = tokens[0]
            gene = tokens[1]
            variant = tokens[2]
        else:
            raise Exception(f'Invalid number of items in hgvs_id=[{hgvs_id}].'
                            f' Should be in the form [{prefix}reference:gene:variant] or [{prefix}reference:variant].')

        return sequence_name, gene, variant

    @classmethod
    def create_from_id(cls, hgvs_id: str, prefix: str = None) -> QueryFeatureHGVS:
        """
        Creates a new HGVS feature. The identifier is in the format hgvs:[sequence]:[gene]:[variant] or
        hgvs:[sequence]:[variant]. This roughly corresponds to the HGVS format <http://varnomen.hgvs.org/>.
        Specifically, the [gene]:[variant] (or [reference]:[variant]) part corresponds to the HGVS identifier.
        I prefix the identifier with the string 'hgvs:' to indicate it is an HGVS identifier. I also include
        the reference genome name in addition to the gene name in cases where the variant is given in gene coordinates.
        :param hgvs_id: The (modified) HGVS identifier.
        :param prefix: The prefix to use for this identifier. Used to override prefix for HGVSGN
        :return: A new HGVS query feature.
        """
        sequence_name, gene, variant = cls.split_id(hgvs_id, prefix=prefix)
        return QueryFeatureHGVS(sequence_name=sequence_name, gene=gene, variant=variant)



from os import path, listdir
from pathlib import Path
from typing import List

import pytest

from genomics_data_index.test.integration import data_dir
from genomics_data_index.test.integration import data_dir_empty


@pytest.fixture
def sample_dirs() -> List[Path]:
    return [data_dir / d for d in listdir(data_dir) if path.isdir(data_dir / d)]


@pytest.fixture
def sample_dirs_empty() -> List[Path]:
    return [data_dir_empty / d for d in listdir(data_dir_empty) if path.isdir(data_dir_empty / d)]



import logging
from typing import Union, List

from ete3 import Tree

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl.TreeSamplesQuery import TreeSamplesQuery
from genomics_data_index.api.query.impl.WrappedSamplesQuery import WrappedSamplesQuery
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet

logger = logging.getLogger(__name__)


class MutationTreeSamplesQuery(TreeSamplesQuery):
    """
    A MutationTreeSamplesQuery represents a query with a tree derived from nucleotide mutation data.
    """

    DISTANCE_UNITS = ['substitutions', 'substitutions/site']

    def __init__(self, connection: DataIndexConnection, wrapped_query: SamplesQuery, tree: Tree,
                 alignment_length: int, reference_name: str, reference_included: bool):
        """
        Builds a new MutationTreeSamplesQuery from the given information. In most normal operations SamplesQuery objects
        are not created directly but are instead created from an :py:class:`genomics_data_index.api.GenomicsDataIndex`
        object or from operations applied to a SamplesQuery (e.g., build_tree() or join_tree()).

        :param connection: A connection to a database containing samples.
        :param wrapped_query: The SamplesQuery to wrap around and join to the passed tree.
        :param tree: The tree to join to this query.
        :param alignment_length: The length of the alignment of mutations used to generate the tree
                                 (used to convert substitutions/site distances to substitutions distances).
        :param reference_name: The name of the reference genome in the tree.
        :param reference_included: True if the reference genome is one of the leaves of the tree, False otherwise.
        :return: A new MutationTreeSamplesQuery object.
        """
        super().__init__(connection=connection, wrapped_query=wrapped_query, tree=tree)
        self._alignment_length = alignment_length
        self._reference_name = reference_name
        self._reference_included = reference_included

    @property
    def reference_name(self) -> str:
        """
        Gets the name of the reference genome.
        :return: The name of the reference genome.
        """
        return self._reference_name

    @property
    def reference_included(self) -> bool:
        """
        Whether or not the reference genome is included as a leaf of the tree.
        :return: True if the reference genome is a leaf of the tree, False otherwise.
        """
        return self._reference_included

    def _get_samples_to_keep_from_query(self, query: SamplesQuery,
                                        include_present: bool,
                                        include_unknown: bool,
                                        include_absent: bool,
                                        ) -> List[str]:
        sample_names_from_query = super()._get_samples_to_keep_from_query(query=query,
                                                                          include_present=include_present,
                                                                          include_unknown=include_unknown,
                                                                          include_absent=include_absent)
        if self.reference_included:
            return sample_names_from_query + [self.reference_name]
        else:
            return sample_names_from_query

    def _create_from_tree_internal(self, tree: Tree) -> SamplesQuery:
        return MutationTreeSamplesQuery(connection=self._query_connection,
                                        wrapped_query=self._wrapped_query,
                                        tree=tree,
                                        alignment_length=self._alignment_length,
                                        reference_name=self.reference_name,
                                        reference_included=self.reference_included)

    def _wrap_create(self, wrapped_query: SamplesQuery, universe_set: SampleSet = None) -> WrappedSamplesQuery:
        return MutationTreeSamplesQuery(connection=self._query_connection,
                                        wrapped_query=wrapped_query,
                                        tree=self._tree,
                                        alignment_length=self._alignment_length,
                                        reference_name=self._reference_name,
                                        reference_included=self._reference_included)

    def _within_distance_internal(self, data: Union[str, List[str], SamplesQuery, SampleSet], distance: float,
                                  units: str) -> SamplesQuery:
        if units == 'substitutions':
            distance_multiplier = self._alignment_length
        elif units == 'substitutions/site':
            distance_multiplier = 1
        else:
            raise Exception(f'Invalid units=[{units}]. Must be one of {self.DISTANCE_UNITS}')

        sample_names, query_infix = self._get_sample_names_query_infix_from_data(data)

        if len(sample_names) == 0:
            found_samples_set = SampleSet.create_empty()
        else:
            sample_name_ids_self = self._get_sample_name_ids(include_unknowns=True)
            found_samples_set = set()
            for sample_name in sample_names:
                sample_leaves = self._tree.get_leaves_by_name(sample_name)
                if len(sample_leaves) != 1:
                    raise Exception(
                        f'Invalid number of matching leaves for sample [{data}], leaves {sample_leaves}')

                sample_node = sample_leaves[0]

                for leaf in self._tree.iter_leaves():
                    if leaf.name not in sample_name_ids_self:
                        continue
                    sample_distance_to_other_sample = sample_node.get_distance(leaf) * distance_multiplier

                    if sample_distance_to_other_sample <= distance:
                        found_samples_set.add(sample_name_ids_self[leaf.name])
        found_samples = SampleSet(found_samples_set)
        return self.intersect(found_samples, f'within({distance} {units} of {query_infix})')

    def _can_handle_distance_units(self, units: str) -> bool:
        return units in self.DISTANCE_UNITS

    def _distance_units(self) -> List[str]:
        return self.DISTANCE_UNITS + self._wrapped_query._distance_units()



from __future__ import annotations

from pathlib import Path
from typing import Union, Tuple, Optional

from ete3 import Tree
from sqlalchemy import Column, String, Integer, LargeBinary, UnicodeText, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model import NUCLEOTIDE_UNKNOWN_TYPE, MLST_UNKNOWN_ALLELE
from genomics_data_index.storage.model.NucleotideMutationTranslater import NucleotideMutationTranslater
from genomics_data_index.storage.model.QueryFeatureMLST import QueryFeatureMLST
from genomics_data_index.storage.model.db.DatabasePathTranslator import DatabasePathTranslator

Base = declarative_base()

# Max of 500 million bytes
MAX_SAMPLE_SET_BYTES = 500 * 10 ** 6

# Used to translate between relative and absolute Paths when persisting to the database.
# TODO: I don't like the idea of using a global variable here and would rather use something more in tune with
# SQLAlchemy (e.g., some hook) or have a manager class keep track of the data root directory. But I haven't
# had time to look into implementing that yet.
# This variable is set in the DatabaseConnection class elsewhere in my code
database_path_translator: Optional[DatabasePathTranslator] = None


class FeatureSamples:
    """
    An abstract class to track properties/methods common to any storage of a link between a feature and
    a set of sample identifiers.

    This is not declared as an abstract class (using abc.ABC) since this was causing metaclass conflicts
    with SQLAlchemy objects below (and I wasn't sure how to resolve them).
    """

    def __init__(self):
        pass

    @hybrid_property
    def id(self) -> str:
        raise NotImplementedError()

    @hybrid_property
    def sample_ids(self) -> SampleSet:
        raise NotImplementedError()

    @sample_ids.setter
    def sample_ids(self, sample_ids: SampleSet) -> None:
        raise NotImplementedError()

    def query_id(self) -> str:
        raise NotImplementedError()

    @hybrid_property
    def is_unknown(self) -> bool:
        raise NotImplementedError()

    def update_sample_ids(self, other: FeatureSamples) -> None:
        """
        Updates (merges) the other objects samples into this object.
        :param other: The other object.
        :return: None. Modifies this object in-place.
        """
        if not isinstance(other, self.__class__):
            raise Exception(f'Cannot merge other=[{other}] since it is not of type {self.__class__}')
        elif self.id != other.id:
            raise Exception(f'Cannot merge other=[{other}] since identifiers are not equal ({self.id} != {other.id}')
        else:
            self.sample_ids = self.sample_ids.union(other.sample_ids)


class NucleotideVariantsSamples(Base, FeatureSamples):
    __tablename__ = 'nucleotide_variants_samples'
    sequence = Column(String(255), primary_key=True)
    position = Column(Integer, primary_key=True)
    deletion = Column(Integer, primary_key=True)
    insertion = Column(String(255), primary_key=True)
    _spdi = Column('spdi', String(255))
    var_type = Column(String(255))
    _sample_ids = Column(LargeBinary(length=MAX_SAMPLE_SET_BYTES))

    # Annotation/effect information
    annotation = Column(String(255))
    annotation_impact = Column(String(255))
    annotation_gene_name = Column(String(255))
    annotation_gene_id = Column(String(255))
    annotation_feature_type = Column(String(255))
    annotation_transcript_biotype = Column(String(255))
    annotation_hgvs_c = Column(String(255))
    annotation_hgvs_p = Column(String(255))

    _id_hgvs_c = Column('id_hgvs_c', String(255))
    _id_hgvs_p = Column('id_hgvs_p', String(255))

    def __init__(self, spdi: str = None, var_type: str = None, sample_ids: SampleSet = None,
                 annotation: str = None, annotation_impact: str = None, annotation_gene_name: str = None,
                 annotation_gene_id: str = None, annotation_feature_type: str = None,
                 annotation_transcript_biotype: str = None, annotation_hgvs_c: str = None,
                 annotation_hgvs_p: str = None):
        super().__init__()
        self.spdi = spdi
        self.var_type = var_type
        self.sample_ids = sample_ids

        self.annotation = annotation
        self.annotation_impact = annotation_impact
        self.annotation_gene_name = annotation_gene_name
        self.annotation_gene_id = annotation_gene_id
        self.annotation_feature_type = annotation_feature_type
        self.annotation_transcript_biotype = annotation_transcript_biotype
        self.annotation_hgvs_c = annotation_hgvs_c
        self.annotation_hgvs_p = annotation_hgvs_p

        self._id_hgvs_c = self.id_hgvs_c
        self._id_hgvs_p = self.id_hgvs_p

    @property
    def id(self) -> str:
        return self.spdi

    @property
    def is_unknown(self) -> bool:
        return self.var_type == NUCLEOTIDE_UNKNOWN_TYPE

    @property
    def query_id(self) -> str:
        return self.id

    @hybrid_property
    def spdi(self) -> str:
        return self.to_spdi(self.sequence, self.position, self.deletion, self.insertion)

    @spdi.setter
    def spdi(self, spdi_value: str):
        self.sequence, self.position, self.deletion, self.insertion = self.from_spdi(spdi_value)
        self._spdi = self.to_spdi(self.sequence, self.position, self.deletion, self.insertion)

    @hybrid_property
    def id_hgvs_c(self) -> str:
        return self.to_hgvs(self.sequence, self.annotation_gene_id, self.annotation_hgvs_c)

    @hybrid_property
    def id_hgvs_p(self) -> str:
        return self.to_hgvs(self.sequence, self.annotation_gene_id, self.annotation_hgvs_p)

    @hybrid_property
    def sample_ids(self) -> SampleSet:
        if self._sample_ids is None:
            raise Exception('_sample_ids is not set')
        else:
            return SampleSet.from_bytes(self._sample_ids)

    @sample_ids.setter
    def sample_ids(self, sample_ids: SampleSet) -> None:
        if sample_ids is None:
            raise Exception('Cannot set sample_ids to None')
        else:
            self._sample_ids = sample_ids.get_bytes()

    @classmethod
    def to_spdi(cls, sequence_name: str, position: int, ref: Union[str, int], alt: str) -> str:
        return NucleotideMutationTranslater.to_spdi(sequence_name, position, ref, alt)

    @classmethod
    def to_hgvs(cls, sequence_name: str, gene_id: Optional[str], variant: str) -> str:
        return NucleotideMutationTranslater.to_hgvs_id(sequence_name=sequence_name,
                                                       gene=gene_id,
                                                       variant=variant)

    @classmethod
    def from_spdi(cls, spdi: str) -> Tuple[str, int, int, str]:
        return NucleotideMutationTranslater.from_spdi(spdi)

    def __repr__(self):
        return (
            f'<NucleotideVariantsSamples(spdi={self.spdi}, var_type={self.var_type}, num_samples={len(self.sample_ids)})>')


class Reference(Base):
    __tablename__ = 'reference'
    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    length = Column(Integer)
    _tree = Column('tree', UnicodeText(10 ** 6))
    tree_alignment_length = Column(Integer)

    sequences = relationship('ReferenceSequence')
    sample_nucleotide_variation = relationship('SampleNucleotideVariation', back_populates='reference')

    def has_tree(self) -> bool:
        return self._tree is not None

    @hybrid_property
    def tree(self) -> Tree:
        if self._tree is None:
            raise Exception('Cannot convert an empty tree')
        else:
            return Tree(self._tree)

    @tree.setter
    def tree(self, in_tree: Tree):
        if in_tree is None:
            self._tree = None
        else:
            self._tree = in_tree.write()

    def __repr__(self):
        return f'<Reference(id={self.id}, name={self.name}, length={self.length})>'


class SampleNucleotideVariation(Base):
    __tablename__ = 'sample_nucleotide_variation'
    sample_id = Column(Integer, ForeignKey('sample.id'), primary_key=True)
    reference_id = Column(Integer, ForeignKey('reference.id'), primary_key=True)
    _masked_regions_file = Column('masked_regions_file', String(255))
    _nucleotide_variants_file = Column('nucleotide_variants_file', String(255))

    @hybrid_property
    def nucleotide_variants_file(self) -> Path:
        if self._nucleotide_variants_file is None:
            raise Exception('Empty _nucleotide_variants_file')
        else:
            if database_path_translator is None:
                raise Exception('Empty database_path_translator')
            else:
                return database_path_translator.from_database(self._nucleotide_variants_file)

    @nucleotide_variants_file.setter
    def nucleotide_variants_file(self, file: Path) -> None:
        if file is None:
            self._nucleotide_variants_file = None
        else:
            if database_path_translator is None:
                raise Exception('Empty database_path_translator')
            else:
                self._nucleotide_variants_file = database_path_translator.to_database(file)

    @property
    def masked_regions(self) -> MaskedGenomicRegions:
        return MaskedGenomicRegions.from_file(self.masked_regions_file)

    @hybrid_property
    def masked_regions_file(self) -> Path:
        if self._masked_regions_file is None:
            raise Exception('Empty _masked_regions_file')
        else:
            if database_path_translator is None:
                raise Exception('Empty database_path_translator')
            else:
                return database_path_translator.from_database(self._masked_regions_file)

    @masked_regions_file.setter
    def masked_regions_file(self, file: Path) -> None:
        if file is None:
            self._masked_regions_file = None
        else:
            if database_path_translator is None:
                raise Exception('Empty database_path_translator')
            else:
                self._masked_regions_file = database_path_translator.to_database(file)

    sample = relationship('Sample', back_populates='sample_nucleotide_variation')
    reference = relationship('Reference', back_populates='sample_nucleotide_variation')


class ReferenceSequence(Base):
    __tablename__ = 'reference_sequence'
    id = Column(Integer, primary_key=True)
    reference_id = Column(Integer, ForeignKey('reference.id'))
    sequence_name = Column(String(255))
    sequence_length = Column(Integer)

    def __repr__(self):
        return (f'<ReferenceSequence(id={self.id}, sequence_name={self.sequence_name},'
                f'sequence_length={self.sequence_length}, reference_id={self.reference_id})>')


class MLSTScheme(Base):
    __tablename__ = 'mlst_scheme'
    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    alleles_dir = Column(String(255))
    sequence_types_file = Column(String(255))

    def __repr__(self):
        return (f'<MLSTScheme(id={self.id}, name={self.name}, '
                f'alleles_dir={self.alleles_dir}, sequence_types_file={self.sequence_types_file})>')


class SampleMLSTAlleles(Base):
    __tablename__ = 'sample_mlst_alleles'
    sample_id = Column(Integer, ForeignKey('sample.id'), primary_key=True)
    scheme_id = Column(Integer, ForeignKey('mlst_scheme.id'), primary_key=True)
    _alleles_file = Column('alleles_file', String(255))

    @hybrid_property
    def alleles_file(self) -> Path:
        if self._alleles_file is None:
            raise Exception('Empty _alleles_file')
        else:
            if database_path_translator is None:
                raise Exception('Empty database_path_translator')
            else:
                return database_path_translator.from_database(self._alleles_file)

    @alleles_file.setter
    def alleles_file(self, file: Path) -> None:
        if file is None:
            self._alleles_file = None
        else:
            if database_path_translator is None:
                raise Exception('Empty database_path_translator')
            else:
                self._alleles_file = database_path_translator.to_database(file)

    sample = relationship('Sample', back_populates='sample_mlst_alleles')
    scheme = relationship('MLSTScheme')

    def __repr__(self):
        return f'<SampleMLSTAlleles(sample_id={self.sample_id}, scheme_id={self.scheme_id}, ' \
               f'_alleles_file={self._alleles_file})>'


class MLSTAllelesSamples(Base, FeatureSamples):
    __tablename__ = 'mlst_alleles_samples'
    scheme = Column(String(255), primary_key=True)
    locus = Column(String(255), primary_key=True)
    allele = Column(String(255), primary_key=True)
    _sla = Column('sla', String(255))
    _sample_ids = Column(LargeBinary(length=MAX_SAMPLE_SET_BYTES))

    def __init__(self, sla: str = None, sample_ids: SampleSet = None):
        super().__init__()
        self.sla = sla
        self.sample_ids = sample_ids

    @hybrid_property
    def id(self) -> str:
        return self.sla

    @property
    def is_unknown(self) -> bool:
        return self.allele == MLST_UNKNOWN_ALLELE

    @property
    def query_id(self) -> str:
        return QueryFeatureMLST.PREFIX + self.id

    @hybrid_property
    def sample_ids(self) -> SampleSet:
        if self._sample_ids is None:
            raise Exception('_sample_ids is not set')
        else:
            return SampleSet.from_bytes(self._sample_ids)

    @sample_ids.setter
    def sample_ids(self, sample_ids: SampleSet) -> None:
        if sample_ids is None:
            raise Exception('Cannot set sample_ids to None')
        else:
            self._sample_ids = sample_ids.get_bytes()

    @hybrid_property
    def sla(self) -> str:
        return self.to_sla(self.scheme, self.locus, self.allele)

    @sla.setter
    def sla(self, sla_value: str):
        self.scheme, self.locus, self.allele = self.from_sla(sla_value)
        self._sla = sla_value

    @classmethod
    def to_sla(cls, scheme_name: str, locus: str, allele: str) -> str:
        return f'{scheme_name}:{locus}:{allele}'

    @classmethod
    def from_sla(cls, sla: str) -> Tuple[str, str, str]:
        if sla is None:
            raise Exception('Cannot parse value sla=None')

        values = sla.split(':')
        if len(values) != 3:
            raise Exception(f'Incorrect number of items for sla=[{sla}]')
        else:
            return str(values[0]), str(values[1]), str(values[2])


class Sample(Base):
    __tablename__ = 'sample'
    id = Column(Integer, primary_key=True)
    name = Column(String(255))

    sample_nucleotide_variation = relationship('SampleNucleotideVariation', back_populates='sample')
    sample_mlst_alleles = relationship('SampleMLSTAlleles', back_populates='sample')
    sample_kmer_index = relationship('SampleKmerIndex', uselist=False, back_populates='sample')

    def __repr__(self):
        return f'<Sample(id={self.id}, name={self.name})>'


class SampleKmerIndex(Base):
    __tablename__ = 'sample_kmer_index'
    sample_id = Column(Integer, ForeignKey('sample.id'), primary_key=True)
    _kmer_index_path = Column('kmer_index_path', String(255))

    sample = relationship('Sample', uselist=False, back_populates='sample_kmer_index')

    def __init__(self, sample: Sample, kmer_index_path: Union[str, Path]):
        self.sample = sample
        self.kmer_index_path = kmer_index_path

    @hybrid_property
    def kmer_index_path(self) -> Path:
        if database_path_translator is None:
            raise Exception('Empty database_path_translator')
        else:
            return database_path_translator.from_database(self._kmer_index_path)

    @kmer_index_path.setter
    def kmer_index_path(self, file: Union[str, Path]):
        if isinstance(file, str):
            file = Path(file)
        if database_path_translator is None:
            raise Exception('Empty database_path_translator')
        else:
            self._kmer_index_path = database_path_translator.to_database(file)



import abc
import logging
from pathlib import Path
from typing import List, Set, Any, Dict, Optional, Generator

import pandas as pd

from genomics_data_index.storage.io.FeaturesReader import FeaturesReader
from genomics_data_index.storage.io.SampleData import SampleData
from genomics_data_index.storage.io.SampleDataPackage import SampleDataPackage
from genomics_data_index.storage.model.db import Sample, FeatureSamples
from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service import EntityExistsError
from genomics_data_index.storage.service.SampleService import SampleService
from genomics_data_index.storage.util import TRACE_LEVEL
from genomics_data_index.storage.util.SamplesProgressLogger import SamplesProgressLogger

logger = logging.getLogger(__name__)

AUTO_SCOPE = '__AUTO_SCOPE__'


class FeatureService(abc.ABC):

    def __init__(self, database_connection: DatabaseConnection, features_dir: Path, sample_service: SampleService,
                 max_insert_batch_size: int = 500):
        self._connection = database_connection
        self._sample_service = sample_service
        self._features_dir = features_dir
        self._max_insert_batch_size = max_insert_batch_size
        self._min_insert_batch_size = 5

    @abc.abstractmethod
    def read_index(self, feature_ids: List[str]) -> Dict[str, FeatureSamples]:
        """
        Reads the passed list of feature IDs from the database and returns a dictionary containing those features that
        already exist. This dictionary maps {'id' => 'feature_object_in_database'}.
        :param feature_ids: The list of feature object ids to search through.
        :return: A dictionary containing any features that already existing in the database index.
        """
        pass

    @abc.abstractmethod
    def get_correct_data_package(self) -> Any:
        pass

    @abc.abstractmethod
    def get_correct_sample_data(self) -> Any:
        pass

    @abc.abstractmethod
    def _create_feature_identifier(self, features_df: pd.DataFrame) -> str:
        pass

    @abc.abstractmethod
    def _get_sample_id_series(self, features_df: pd.DataFrame, sample_name_ids: Dict[str, int]) -> pd.Series:
        pass

    @abc.abstractmethod
    def _create_feature_object(self, features_df: pd.DataFrame) -> FeatureSamples:
        pass

    @abc.abstractmethod
    def aggregate_feature_column(self) -> Dict[str, Any]:
        pass

    def _modify_df_types(self, features_df: pd.DataFrame) -> pd.DataFrame:
        return features_df

    def _create_feature_objects(self, features_df: pd.DataFrame, sample_names: Set[str]) -> List[FeatureSamples]:
        logger.debug(f'Aggregating samples by feature identifiers within a table of {len(features_df)} rows'
                     f' across {len(sample_names)} samples')

        sample_name_ids = self._sample_service.find_sample_name_ids(sample_names)

        features_df['_FEATURE_ID'] = features_df.apply(self._create_feature_identifier, axis='columns')
        features_df['_SAMPLE_ID'] = self._get_sample_id_series(features_df, sample_name_ids)

        index_df = features_df.groupby('_FEATURE_ID').agg(self.aggregate_feature_column())
        index_df = self._modify_df_types(index_df).reset_index()

        feature_objects_list = index_df.apply(self._create_feature_object, axis='columns').tolist()

        logger.debug(f'Finished aggregating {len(sample_names)} samples into a table of {len(index_df)} features '
                     f'from an original table size of {len(features_df)} rows.')
        return feature_objects_list

    @abc.abstractmethod
    def check_samples_have_features(self, sample_names: Set[str], feature_scope_name: str) -> bool:
        """
        Checks if any of the passed sample names already have features.
        :param sample_names: The set of sample names.
        :param feature_scope_name: A name defining the feature scope (i.e., reference genome name).
        :return: True if any of the samples have features, false otherwise.
        """
        pass

    def _verify_correct_data_package(self, data_package: SampleDataPackage) -> None:
        if not isinstance(data_package, self.get_correct_data_package()):
            raise Exception(f'data_package=[{data_package}] is not of type'
                            f' {self.get_correct_data_package().__name__}')

    def _verify_correct_sample_data(self, sample_data: SampleData) -> None:
        if not isinstance(sample_data, self.get_correct_sample_data()):
            raise Exception(f'sample_data=[{sample_data}] is not of type'
                            f' {self.get_correct_sample_data().__name__}')

    def _get_or_create_sample(self, sample_name: str) -> Sample:
        if self._sample_service.exists(sample_name):
            return self._sample_service.get_sample(sample_name)
        else:
            return Sample(name=sample_name)

    def _get_or_create_samples(self, sample_names: List[str]) -> Dict[str, Sample]:
        existing_samples = self._sample_service.get_existing_samples_by_names(sample_names)
        samples_dict = {sample.name: sample for sample in existing_samples}

        for name in sample_names:
            if name not in samples_dict:
                samples_dict[name] = Sample(name=name)

        return samples_dict

    def _verify_correct_feature_scope(self, feature_scope_name: str) -> None:
        if feature_scope_name is None:
            raise Exception('feature_scope_name cannot be None')

    def _set_batch_size(self, num_samples: int) -> int:
        batch_size = max(self._min_insert_batch_size, int(num_samples / 50))
        batch_size = min(self._max_insert_batch_size, batch_size)
        return batch_size

    def _sample_data_iter_batch(self, data_package: SampleDataPackage,
                                batch_size: int) -> Generator[Dict[str, SampleData], None, None]:
        sample_data_batch = {}
        for sample_data in data_package.iter_sample_data():
            sample_data_batch[sample_data.sample_name] = sample_data

            if len(sample_data_batch) >= batch_size:
                yield sample_data_batch
                sample_data_batch = {}

        if len(sample_data_batch) > 0:
            yield sample_data_batch

    def insert(self, data_package: SampleDataPackage, feature_scope_name: str = AUTO_SCOPE) -> None:
        self._verify_correct_data_package(data_package=data_package)
        self._verify_correct_feature_scope(feature_scope_name)

        sample_names = data_package.sample_names()
        num_samples = len(sample_names)
        progress_logger = SamplesProgressLogger(stage_name='Insert', stage_number=1, total_samples=num_samples)

        if self.check_samples_have_features(sample_names, feature_scope_name):
            raise EntityExistsError(f'Passed samples already have features for feature scope [{feature_scope_name}], '
                                    f'will not insert any new features')

        batch_size = self._set_batch_size(num_samples)
        logger.debug(f'Batch size {batch_size}')
        processed_samples = 0
        persisted_sample_data_dict = {}
        progress_logger.update_progress(processed_samples)
        for sample_data_batch in self._sample_data_iter_batch(data_package, batch_size=batch_size):
            persisted_sample_data = self._handle_batch(sample_data_batch, feature_scope_name)
            persisted_sample_data_dict.update(persisted_sample_data)
            processed_samples += len(sample_data_batch)
            progress_logger.update_progress(processed_samples)

        self._connection.get_session().commit()
        logger.info(f'Finished processing {num_samples} samples')

        persisted_features_reader = self._create_persisted_features_reader(sample_data_dict=persisted_sample_data_dict,
                                                                           data_package=data_package)
        self.index_features(features_reader=persisted_features_reader, feature_scope_name=feature_scope_name)

    def _handle_batch(self, sample_data_batch: Dict[str, SampleData], feature_scope_name: str) -> Dict[str, SampleData]:
        samples_dict = self._get_or_create_samples(list(sample_data_batch.keys()))
        persisted_sample_data_dict = {}
        for sample_name in samples_dict:
            logger.log(TRACE_LEVEL, f'Persisting sample {sample_name}')

            sample = samples_dict[sample_name]
            persisted_sample_data = self._persist_sample_data(sample_data_batch[sample.name])
            sample_feature_object = self.build_sample_feature_object(sample=sample, sample_data=persisted_sample_data,
                                                                     feature_scope_name=feature_scope_name)

            persisted_sample_data_dict[persisted_sample_data.sample_name] = persisted_sample_data
            self._connection.get_session().add(sample_feature_object)

        return persisted_sample_data_dict

    def _update_scope(self, features_df: pd.DataFrame, feature_scope_name: str) -> pd.DataFrame:
        return features_df

    def update_feature_objects_inplace(self, feature_objects_dict_dest: Dict[str, FeatureSamples],
                                       feature_objects_dict_source: Dict[str, FeatureSamples]) -> None:
        """
        Updates matching features from feature_objects_dict_source into feature_objects_dict_dest.
        This will modify objects in feature_objects_dict_dest. This will *not* insert any new objects from
        feature_objects_dict_source. This method is primarily used to update the existing objects in a database.
        :param feature_objects_dict_dest: The dictionary of feature ids to feature objects to merge into.
        :param feature_objects_dict_source: The dictionary of feature ids to feature objects to merge from.
        :return: Nothing. Modifies objects in feature_objects_dict_dest.
        """
        logger.debug(f'Scanning {len(feature_objects_dict_dest)} destination features for corresponding'
                     f' source features and updating if any matches are found.')
        count_updated_features = 0
        for feature_id in feature_objects_dict_dest:
            feature_object_dest = feature_objects_dict_dest[feature_id]
            if feature_id in feature_objects_dict_source:
                feature_object_source = feature_objects_dict_source[feature_id]
                feature_object_dest.update_sample_ids(feature_object_source)
                count_updated_features += 1

        logger.debug(f'Updated {count_updated_features}/{len(feature_objects_dict_dest)} destination feature objects '
                     f'inplace out of a total of {len(feature_objects_dict_source)} supplied source feature objects.')

    def index_features(self, features_reader: FeaturesReader, feature_scope_name: str) -> None:
        logger.info(f'Reading features from {len(features_reader.samples_list())} samples')
        features_df = features_reader.get_features_table()
        features_df = self._update_scope(features_df, feature_scope_name)
        sample_names = features_reader.samples_set()

        # Create new features and merge with existing features if they exist
        logger.info(f'Aggregating {len(features_df)} features found in {len(sample_names)} samples')
        created_feature_objects = self._create_feature_objects(features_df, sample_names)
        created_feature_objects_dict = {f.id: f for f in created_feature_objects}
        logger.info(f'A total of {len(created_feature_objects_dict)} unique features across {len(sample_names)} '
                    f'samples will be updated. Searching for existing features in the database.')
        db_feature_objects_dict = self.read_index(list(created_feature_objects_dict.keys()))
        logger.info(f'Found {len(db_feature_objects_dict)}/{len(created_feature_objects_dict)} features in database '
                    f'which already exist. These features will be updated to map to the new samples.')
        self.update_feature_objects_inplace(db_feature_objects_dict, created_feature_objects_dict)

        # Subtract out existing feature objects in database
        new_feature_object_ids = set(created_feature_objects_dict.keys()) - set(db_feature_objects_dict.keys())
        new_feature_objects = [created_feature_objects_dict[fid] for fid in new_feature_object_ids]

        # bulk update existing feature objects
        logger.info(
            f'Updating {len(db_feature_objects_dict)}/{len(created_feature_objects_dict)} existing features in database')
        self._connection.get_session().bulk_save_objects(db_feature_objects_dict.values())

        # bulk save new feature objects
        logger.info(
            f'Saving {len(new_feature_objects)}/{len(created_feature_objects_dict)} new features in database')
        self._connection.get_session().bulk_save_objects(new_feature_objects)

        self._connection.get_session().commit()
        logger.info(f'Finished indexing features from {len(sample_names)} samples')

    @abc.abstractmethod
    def build_sample_feature_object(self, sample: Sample, sample_data: SampleData, feature_scope_name: str) -> Any:
        pass

    def _persist_sample_data(self, sample_data: SampleData) -> Optional[SampleData]:
        if sample_data is not None:
            return sample_data.persist(self._features_dir)
        else:
            return None

    @abc.abstractmethod
    def _create_persisted_features_reader(self, sample_data_dict: Dict[str, SampleData],
                                          data_package: SampleDataPackage) -> FeaturesReader:
        pass



from pathlib import Path
from typing import Dict


class ExecutorResults:

    def __init__(self, files: Dict[str, Path]):
        self._files = files

    def get_file(self, key: str) -> Path:
        return self._files[key]



import logging
from typing import List, Dict, Set, Union, cast, Type

import pandas as pd

from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model.NucleotideMutationTranslater import NucleotideMutationTranslater
from genomics_data_index.storage.model.QueryFeature import QueryFeature
from genomics_data_index.storage.model.QueryFeatureHGVS import QueryFeatureHGVS
from genomics_data_index.storage.model.QueryFeatureHGVSGN import QueryFeatureHGVSGN
from genomics_data_index.storage.model.QueryFeatureMLST import QueryFeatureMLST
from genomics_data_index.storage.model.QueryFeatureMutation import QueryFeatureMutation
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI
from genomics_data_index.storage.model.db import NucleotideVariantsSamples, Reference, ReferenceSequence, MLSTScheme, \
    SampleMLSTAlleles, MLSTAllelesSamples, Sample
from genomics_data_index.storage.model.db import SampleNucleotideVariation
from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service import SQLQueryInBatcherDict, SQLQueryInBatcherList

logger = logging.getLogger(__name__)


class FeatureExplodeUnknownError(Exception):

    def __init__(self, msg: str):
        super().__init__(msg)


class SampleService:

    def __init__(self, database_connection: DatabaseConnection, sql_select_limit: int):
        self._connection = database_connection
        self._sql_select_limit = sql_select_limit

    def get_samples_with_variants(self, reference_name: str) -> List[Sample]:
        """
        Gets a list of all samples that have variants associated with the given reference genome name.
        :reference_name: The reference genome name.
        :return: A list of Samples with variants with respect to the reference genome name, empty list of no Samples.
        """
        samples = self._connection.get_session().query(Sample) \
            .join(Sample.sample_nucleotide_variation) \
            .join(SampleNucleotideVariation.reference) \
            .filter(Reference.name == reference_name) \
            .all()
        return samples

    def feature_explode_unknown(self, feature: QueryFeature) -> List[QueryFeature]:
        if isinstance(feature, QueryFeatureHGVSGN):
            features_spdi = self.find_features_spdi_for_hgvsgn(feature)

            if len(features_spdi) == 0:
                raise FeatureExplodeUnknownError(f'feature={feature} is of type HGVSGN but the corresponding SPDI '
                                                 f'feature does not exist in the database. Cannot convert to unknown '
                                                 f'SPDI representation.')
            else:
                unknown_features = []
                for feature in features_spdi:
                    unknown_features.extend(feature.to_unknown_explode())
                return unknown_features
        elif isinstance(feature, QueryFeatureHGVS):
            if feature.is_nucleotide():
                variants_hgvs = self._connection.get_session().query(NucleotideVariantsSamples) \
                    .filter(NucleotideVariantsSamples._id_hgvs_c == feature.id) \
                    .all()
            elif feature.is_protein():
                variants_hgvs = self._connection.get_session().query(NucleotideVariantsSamples) \
                    .filter(NucleotideVariantsSamples._id_hgvs_p == feature.id) \
                    .all()
            else:
                raise Exception(f'feature=[{feature}] is neither nucleotide or protein')

            if len(variants_hgvs) == 0:
                raise FeatureExplodeUnknownError(f'feature={feature} is of type HGVS but the corresponding SPDI '
                                                 f'feature does not exist in the database. Cannot convert to unknown '
                                                 f'SPDI representation.')
            else:
                unknown_features = []
                for variants_sample_obj in variants_hgvs:
                    unknown_features.extend(QueryFeatureMutationSPDI(variants_sample_obj.spdi).to_unknown_explode())

                return unknown_features
        else:
            return feature.to_unknown_explode()

    def find_features_spdi_for_hgvsgn(self, feature: QueryFeatureHGVSGN) -> List[QueryFeatureMutationSPDI]:
        if not isinstance(feature, QueryFeatureHGVSGN):
            raise Exception(f'Cannot handle feature={feature}. Not of type {QueryFeatureHGVSGN.__name__}')

        query = self._connection.get_session().query(NucleotideVariantsSamples).filter(
            NucleotideVariantsSamples.sequence == feature.sequence)

        if feature.has_gene():
            query = query.filter(NucleotideVariantsSamples.annotation_gene_name == feature.gene)

        if feature.is_nucleotide():
            query = query.filter(NucleotideVariantsSamples.annotation_hgvs_c == feature.variant)
        elif feature.is_protein():
            query = query.filter(NucleotideVariantsSamples.annotation_hgvs_p == feature.variant)
        else:
            raise Exception(f'feature={feature} is neither protein nor nucleotide')

        return [QueryFeatureMutationSPDI(s.spdi) for s in query.all()]

    def get_samples_with_mlst_alleles(self, scheme_name: str) -> List[Sample]:
        """
        Gets a list of all samples that have MLST alleles associated with the given scheme name.
        :scheme_name: The scheme name.
        :return: A list of Samples with MLST alleles with respect to the scheme name, empty list of no Samples.
        """
        samples = self._connection.get_session().query(Sample) \
            .join(Sample.sample_mlst_alleles) \
            .join(SampleMLSTAlleles.scheme) \
            .filter(MLSTScheme.name == scheme_name) \
            .all()
        return samples

    def get_samples_with_variants_on_sequence(self, sequence_name: str) -> List[Sample]:
        """
        Gets a list of all samples that have variants associated with the given sequence name.
        :sequence_name: The sequence name.
        :return: A list of Samples with variants with respect to the sequence name, empty list of no Samples.
        """
        samples = self._connection.get_session().query(Sample) \
            .join(Sample.sample_nucleotide_variation) \
            .join(SampleNucleotideVariation.reference) \
            .join(Reference.sequences) \
            .filter(ReferenceSequence.sequence_name == sequence_name) \
            .all()
        return samples

    def get_samples_associated_with_reference(self, reference_name: str) -> List[Sample]:
        """
        Gets a list of all samples associated with a reference name.
        :reference_name: The reference name.
        :return: A list of Samples associated with the reference name or an empty list if no Samples.
        """
        samples = self._connection.get_session().query(Sample) \
            .join(Sample.sample_nucleotide_variation) \
            .join(SampleNucleotideVariation.reference) \
            .filter(Reference.name == reference_name) \
            .all()
        return samples

    def get_samples_set_associated_with_reference(self, reference_name: str) -> SampleSet:
        """
        Gets a list of all samples associated with a reference name.
        :reference_name: The reference name.
        :return: A list of Samples associated with the reference name or an empty list if no Samples.
        """
        sample_ids = [i for i, in self._connection.get_session().query(Sample.id) \
            .join(Sample.sample_nucleotide_variation) \
            .join(SampleNucleotideVariation.reference) \
            .filter(Reference.name == reference_name) \
            .all()]

        return SampleSet(sample_ids=sample_ids)

    def create_dataframe_from_sample_set(self, present_set: SampleSet,
                                         absent_set: SampleSet,
                                         unknown_set: SampleSet,
                                         queries_expression: str) -> pd.DataFrame:
        sample_sets_status_list = [(present_set, 'Present'), (absent_set, 'Absent'), (unknown_set, 'Unknown')]
        data = []
        for sample_status in sample_sets_status_list:
            sample_set = sample_status[0]
            status = sample_status[1]

            if not sample_set.is_empty():
                samples = self.find_samples_by_ids(sample_set)
                for sample in samples:
                    data.append([queries_expression, sample.name, sample.id, status])

        return pd.DataFrame(data=data, columns=['Query', 'Sample Name', 'Sample ID', 'Status'])

    def count_samples_associated_with_reference(self, reference_name: str) -> int:
        return self._connection.get_session().query(Sample) \
            .join(Sample.sample_nucleotide_variation) \
            .join(SampleNucleotideVariation.reference) \
            .filter(Reference.name == reference_name) \
            .count()

    def count_samples_associated_with_mlst_scheme(self, scheme_name: str) -> int:
        return len(self.get_samples_with_mlst_alleles(scheme_name))

    def get_samples(self) -> List[Sample]:
        return self._connection.get_session().query(Sample).all()

    def count_samples(self) -> int:
        return self._connection.get_session().query(Sample).count()

    def get_all_sample_ids(self) -> SampleSet:
        ids_list = [id for id, in self._connection.get_session().query(Sample.id).all()]
        return SampleSet(ids_list)

    def get_existing_samples_by_names(self, sample_names: List[str]) -> List[Sample]:
        return self._connection.get_session().query(Sample) \
            .filter(Sample.name.in_(sample_names)) \
            .all()

    def which_exists(self, sample_names: List[str]) -> List[str]:
        """
        Returns which of the given samples exist in the database.
        :param sample_names: The list of sample names.
        :return: A list of those passed sample names that exist in the database.
        """
        samples = self._connection.get_session().query(Sample) \
            .filter(Sample.name.in_(sample_names)) \
            .all()
        return [sample.name for sample in samples]

    def get_sample(self, sample_name: str) -> Sample:
        return self._connection.get_session().query(Sample) \
            .filter(Sample.name == sample_name) \
            .one()

    def exists(self, sample_name: str):
        return self._connection.get_session().query(Sample) \
                   .filter(Sample.name == sample_name).count() > 0

    def find_samples_by_ids(self, sample_ids: Union[List[int], SampleSet]) -> List[Sample]:
        if isinstance(sample_ids, SampleSet):
            sample_ids = list(sample_ids)

        query_batcher = SQLQueryInBatcherList(in_data=sample_ids, batch_size=self._sql_select_limit)

        def handle_batch(sample_ids_batch: List[int]) -> List[Sample]:
            return self._connection.get_session().query(Sample) \
                .filter(Sample.id.in_(sample_ids_batch)) \
                .all()

        return query_batcher.process(handle_batch)

    def get_variants_samples_by_variation_features(self, features: List[QueryFeatureMutation]) -> Dict[
        str, NucleotideVariantsSamples]:
        standardized_features_to_input_feature = {}
        standardized_features_ids = set()
        standardized_feature_hgvs_c_ids = set()
        standardized_feature_hgvs_p_ids = set()
        for feature in features:
            if isinstance(feature, QueryFeatureMutationSPDI):
                dbf = NucleotideMutationTranslater.to_db_feature(feature)
                if dbf.id in standardized_features_to_input_feature:
                    standardized_features_to_input_feature[dbf.id].append(feature.id)
                else:
                    standardized_features_to_input_feature[dbf.id] = [feature.id]
                standardized_features_ids.add(dbf.id)
            elif isinstance(feature, QueryFeatureHGVSGN):
                logger.warning(f'feature=[{feature}] is a QueryFeatureHGVSGN and I do not handle it here.')
            elif isinstance(feature, QueryFeatureHGVS):
                if feature.is_nucleotide():
                    standardized_feature_hgvs_c_ids.add(feature.id)
                elif feature.is_protein():
                    standardized_feature_hgvs_p_ids.add(feature.id)
                else:
                    raise Exception(f'feature=[{feature}] is neither nucleotide or protein')
            else:
                raise Exception(f'Invalid type for feature=[{feature}]. '
                                f'Must be either {QueryFeatureMutationSPDI.__class__.__name__} or '
                                f'{QueryFeatureHGVS.__class__.__name__}')

        if len(standardized_features_ids) > 0:
            variants_spdi = self._connection.get_session().query(NucleotideVariantsSamples) \
                .filter(NucleotideVariantsSamples._spdi.in_(standardized_features_ids)) \
                .all()
        else:
            variants_spdi = []

        if len(standardized_feature_hgvs_c_ids) > 0:
            variants_hgvs_c = self._connection.get_session().query(NucleotideVariantsSamples) \
                .filter(NucleotideVariantsSamples._id_hgvs_c.in_(standardized_feature_hgvs_c_ids)) \
                .all()
        else:
            variants_hgvs_c = []

        if len(standardized_feature_hgvs_p_ids) > 0:
            variants_hgvs_p = self._connection.get_session().query(NucleotideVariantsSamples) \
                .filter(NucleotideVariantsSamples._id_hgvs_p.in_(standardized_feature_hgvs_p_ids)) \
                .all()
        else:
            variants_hgvs_p = []

        # Map back unstandardized IDs to the actual variant object
        # Use this because some features can have multiple identifiers for the same feature
        # (e.g., ref:10:A:T and ref:10:1:T). I want to make sure I map each passed id to the
        # same object (that is, in this example, I want to return a dictionary with two keys, one for each ID)
        unstandardized_variants = {}
        for v in variants_spdi:
            for vid in standardized_features_to_input_feature[v.spdi]:
                unstandardized_variants[vid] = v

        unstandardized_variants.update({v.id_hgvs_c: v for v in variants_hgvs_c})
        unstandardized_variants.update({v.id_hgvs_p: v for v in variants_hgvs_p})
        return unstandardized_variants

    def _get_mlst_samples_by_mlst_features(self, features: List[QueryFeatureMLST]) -> List[MLSTAllelesSamples]:
        feature_ids = list({f.id_no_prefix for f in features})
        mlst_alleles = self._connection.get_session().query(MLSTAllelesSamples) \
            .filter(MLSTAllelesSamples._sla.in_(feature_ids)) \
            .all()

        return mlst_alleles

    def _get_feature_type(self, features: List[QueryFeature]) -> Type[QueryFeature]:
        feature_types = {f.__class__ for f in features}

        if len(feature_types) != 1:
            raise Exception(f'Should only be one feature type but instead got: {feature_types}.')
        else:
            return feature_types.pop()

    def find_unknown_sample_sets_by_features(self, features: List[QueryFeature]) -> Dict[str, SampleSet]:
        unknown_to_features_dict = {}
        unknown_features = []
        for feature in features:
            try:
                unknown_features_exploded = self.feature_explode_unknown(feature)
                unknown_features.extend(unknown_features_exploded)
                for unknown_feature in unknown_features_exploded:
                    unknown_to_features_dict[unknown_feature.id] = feature
            except FeatureExplodeUnknownError as e:
                logger.warning(
                    f'Could not map feature={feature} to a set of unknown features. Will assume no unknowns exist.')

        if len(unknown_features) > 0:
            unknown_features_sets = self.find_sample_sets_by_features(unknown_features)
        else:
            unknown_features_sets = set()

        features_to_unknown_sample_sets = {}
        for uid in unknown_features_sets:
            fid = unknown_to_features_dict[uid].id
            sample_set = unknown_features_sets[uid]

            # If we've already set this sample set with the same feature,
            # We need to merge together the unknown sample sets
            # This can occur if, e.g., we have a large deletion and are iterating over each
            # Base in the deletion in turn (e.g., ref:10:ATT:A -> gets converted to
            # ['ref:10:A:?', 'ref:11:T:?', 'ref:12:T:?'], we need to merge unknown sample results
            # for each of these features in turn.
            if fid in features_to_unknown_sample_sets:
                previous_sample_set = features_to_unknown_sample_sets[fid]
                features_to_unknown_sample_sets[fid] = previous_sample_set.union(sample_set)
            else:
                features_to_unknown_sample_sets[fid] = sample_set

        return features_to_unknown_sample_sets

    def find_sample_sets_by_features(self, features: List[QueryFeature]) -> Dict[str, SampleSet]:
        feature_type = self._get_feature_type(features)

        if issubclass(feature_type, QueryFeatureHGVSGN):
            # In this case where I'm querying by gene name, first convert to SPDI features before lookup
            # TODO: it's not the most efficient to do this as a loop, but it's easier to implement right now
            hgvs_gn_id_to_sampleset = dict()
            for feature in features:
                feature = cast(QueryFeatureHGVSGN, feature)
                features_spdi = self.find_features_spdi_for_hgvsgn(feature)
                variants_dict = self.get_variants_samples_by_variation_features(features_spdi)
                variants_nuc_variants_samples = list(variants_dict.values())

                if len(variants_nuc_variants_samples) == 0:
                    samples_union = SampleSet.create_empty()
                else:
                    first_nuc_variant_samples = variants_nuc_variants_samples.pop()
                    samples_union = first_nuc_variant_samples.sample_ids

                    # Handle remaining, if any
                    for nuc_variant_samples in variants_nuc_variants_samples:
                        samples_union = samples_union.union(nuc_variant_samples.sample_ids)

                hgvs_gn_id_to_sampleset[feature.id] = samples_union
            return hgvs_gn_id_to_sampleset
        elif issubclass(feature_type, QueryFeatureMutation):
            features = cast(List[QueryFeatureMutation], features)
            variants_dict = self.get_variants_samples_by_variation_features(features)

            return {id: variants_dict[id].sample_ids for id in variants_dict}
        elif issubclass(feature_type, QueryFeatureMLST):
            features = cast(List[QueryFeatureMLST], features)
            mlst_alleles = self._get_mlst_samples_by_mlst_features(features)

            return {a.query_id: a.sample_ids for a in mlst_alleles}
        else:
            raise Exception(f'Invalid feature type {feature_type}')

    def find_samples_by_features(self, features: List[QueryFeature]) -> Dict[str, List[Sample]]:
        feature_type = self._get_feature_type(features)

        if issubclass(feature_type, QueryFeatureMutation):
            features = cast(List[QueryFeatureMutation], features)
            variants_dict = self.get_variants_samples_by_variation_features(features)

            return {id: self.find_samples_by_ids(variants_dict[id].sample_ids) for id in variants_dict}
        elif issubclass(feature_type, QueryFeatureMLST):
            features = cast(List[QueryFeatureMLST], features)
            mlst_alleles = self._get_mlst_samples_by_mlst_features(features)

            return {a.query_id: self.find_samples_by_ids(a.sample_ids) for a in mlst_alleles}
        else:
            raise Exception(f'Invalid feature type {feature_type}')

    def count_samples_by_features(self, features: List[QueryFeature]) -> Dict[str, List[Sample]]:
        feature_type = self._get_feature_type(features)

        if issubclass(feature_type, QueryFeatureMutation):
            features = cast(List[QueryFeatureMutation], features)
            variants_dict = self.get_variants_samples_by_variation_features(features)

            return {id: len(variants_dict[id].sample_ids) for id in variants_dict}
        elif issubclass(feature_type, QueryFeatureMLST):
            features = cast(List[QueryFeatureMLST], features)
            mlst_alleles = self._get_mlst_samples_by_mlst_features(features)

            allele_id_to_count = {a.query_id: len(a.sample_ids) for a in mlst_alleles}
            for f in features:
                if f.id not in allele_id_to_count:
                    allele_id_to_count[f.id] = 0
            return allele_id_to_count
        else:
            raise Exception(f'Invalid feature type {feature_type}')

    def find_sample_name_ids(self, sample_names: Set[str]) -> Dict[str, int]:
        """
        Given a list of sample names, returns a dictionary mapping the sample names to sample IDs.
        :param sample_names: The sample names to search.
        :return: A dictionary linking the sample names to IDs.
        """
        query_batcher = SQLQueryInBatcherDict(in_data=list(sample_names), batch_size=self._sql_select_limit)

        def handle_batch(sample_names_batch: List[str]) -> Dict[str, int]:
            sample_tuples = self._connection.get_session().query(Sample.name, Sample.id) \
                .filter(Sample.name.in_(sample_names_batch)) \
                .all()
            return dict(sample_tuples)

        return query_batcher.process(handle_batch)

    def get_sample_set_by_names(self, sample_names: Union[List[str], Set[str]],
                                ignore_not_found: bool = False) -> SampleSet:
        """
        Given a collection of sample names, get a SampleSet of the corresponding IDs.
        :param sample_names: The names to convert to an ID set.
        :param ignore_not_found: Whether or not to ignore sample names that were not found.
        :return: A SampleSet with all the corresponding samples by the passed names. If ignore_not_found is false,
                 raises an exception if some sample names have no ids.
        """
        if isinstance(sample_names, list):
            sample_names = set(sample_names)
        elif not isinstance(sample_names, set):
            raise Exception(f'Invalid type=[{type(sample_names)}] for passed sample_names. Must be list or set.')

        sample_ids_tuples = self._connection.get_session().query(Sample.id) \
            .filter(Sample.name.in_(sample_names)) \
            .all()
        sample_ids = {i for i, in sample_ids_tuples}
        sample_set = SampleSet(sample_ids=sample_ids)

        if ignore_not_found or len(sample_names) == len(sample_set):
            return sample_set
        else:
            # Find matching sample names to ids we did find for a nicer error message
            found_sample_names = {s.name for s in self.find_samples_by_ids(sample_set)}
            names_not_found = sample_names - found_sample_names
            if len(names_not_found) > 10:
                small_not_found = list(names_not_found)[:10]
                msg = f'[{", ".join(small_not_found)}, ...]'
            else:
                msg = f'{names_not_found}'

            raise Exception(f'Did not find an equal number of sample names and ids. '
                            f'Number sample_names={len(sample_names)}. Number returned sample_ids={len(sample_ids)}. '
                            f'Sample names with missing ids {msg}')



from __future__ import annotations

from typing import List, Union

from genomics_data_index.storage.model.QueryFeature import QueryFeature


class QueriesCollection:

    def __init__(self, queries_list: List[Union[QueryFeature, str]]):
        self._queries = queries_list

    def append(self, query_obj: Union[QueryFeature, str]) -> QueriesCollection:
        new_queries_list = self._queries.copy()
        new_queries_list.append(query_obj)
        return QueriesCollection(new_queries_list)

    @property
    def last(self):
        return self._queries[-1]

    def query_expression(self) -> str:
        query_strs = [str(q) for q in self._queries]
        return ' AND '.join(query_strs)

    @classmethod
    def create_empty(cls):
        return QueriesCollection(queries_list=[])



from __future__ import annotations

import configparser
import logging
import os
import tempfile
from pathlib import Path

from genomics_data_index.storage.util import execute_command

logger = logging.getLogger(__name__)


class SnpEffDatabase:

    def __init__(self, snpeff_config: Path, database_dir: Path, genome_name: str):
        self._snpeff_config = snpeff_config
        self._database_dir = database_dir
        self._genome_name = genome_name

    def write_database_config(self, file: Path) -> None:
        database_config = configparser.ConfigParser()
        database_config['snpeff'] = {
            'snpeff_config': str(self._snpeff_config),
            'database_dir': str(self._database_dir),
            'genome_name': str(self._genome_name),
        }

        logger.debug(f'Writing snpeff database config to [{file}]')
        with open(file, 'w') as fh:
            database_config.write(fh)

    @property
    def config(self) -> Path:
        return self._snpeff_config

    @property
    def genome_name(self) -> str:
        return self._genome_name

    @property
    def database_dir(self) -> Path:
        return self._database_dir

    def annotate(self, input_vcf_file: Path, output_vcf_file: Path) -> Path:
        """
        Annotates the given VCF file with the stored snpeff database.
        :param input_vcf_file: The VCF file to annotate.
        :param output_vcf_file: The output VCF file to write to.
        :return: The annotated VCF file.
        """
        # Create temporary directory to avoid cluttering up location of main VCF file with additional outputs (if any)
        # And to provide a place to output a VCF file and then bgzip it
        with tempfile.TemporaryDirectory() as out_dir:
            tmp_out_dir = Path(out_dir)
            tmp_vcf_path = tmp_out_dir / input_vcf_file.name
            tmp_vcf_out = tmp_out_dir / 'output.vcf'
            os.symlink(input_vcf_file.absolute(), tmp_vcf_path)

            snpeff_command = ['snpEff', 'ann', '-v', '-c', str(self._snpeff_config),
                              '-hgvs', '-hgvs1LetterAa', '-no-upstream', '-no-downstream', '-i', 'vcf', '-o', 'vcf',
                              '-noStats', '-sequenceOntology', '-nodownload', '-noLog',
                              self._genome_name, str(tmp_vcf_path)]

            # Execute snpeff and write data to output_vcf_h
            with open(tmp_vcf_out, 'w') as out_vcf_h:
                execute_command(snpeff_command, stdout=out_vcf_h)

            # compress file with bgzip
            bgzip_command = ['bgzip', '--stdout', str(tmp_vcf_out)]
            with open(output_vcf_file, 'w') as bgzip_out_h:
                execute_command(bgzip_command, stdout=bgzip_out_h)

        return output_vcf_file

    @classmethod
    def create_from_config(cls, config_file: Path) -> SnpEffDatabase:
        database_config = configparser.ConfigParser()
        database_config.read(config_file)

        snpeff_config = database_config['snpeff']['snpeff_config']
        database_dir = database_config['snpeff']['database_dir']
        genome_name = database_config['snpeff']['genome_name']

        if snpeff_config is None or database_dir is None or genome_name is None:
            raise Exception(f'Invalid snpeff database config in [{config_file}]')
        else:
            snpeff_config = Path(snpeff_config)
            database_dir = Path(database_dir)

        return SnpEffDatabase(snpeff_config=snpeff_config,
                              database_dir=database_dir,
                              genome_name=genome_name)



import logging
from pathlib import Path
from typing import Generator, List

from genomics_data_index.storage.io.SampleData import SampleData
from genomics_data_index.storage.io.SampleFilesProcessor import SampleFilesProcessor

logger = logging.getLogger(__name__)


class SerialSampleFilesProcessor(SampleFilesProcessor):

    def __init__(self, preprocess_dir: Path):
        super().__init__()
        self._preprocess_dir = preprocess_dir

    def process(self, sample_data: List[SampleData]) -> Generator[SampleData, None, None]:
        for sample_files in sample_data:
            logger.debug(f'Pre-processing files for sample [{sample_files.sample_name}]')
            yield sample_files.persist(self._preprocess_dir)
